#!/etc/opt/openit/perl

BEGIN {
	if ( $^O ne 'MSWin32' ) {
		push(@INC, "/opt/openit/lib" );
        $ENV{'OPENIT_CFG'} = "/etc/opt/openit/openit.cfg"; 
	} 
	push(@INC, $ENV{'ALTUS'} . "/src/perl/source" ); 
}
## Common Modules
use strict;
use Getopt::Long;
use File::Basename;
use File::Copy;
use Data::Dumper;
use Cwd 'abs_path';

## OpeniT Modules
use OpeniT::Config;
use OpeniT::Defaults;
use OpeniT::DataTypes;
use OpeniT::File;

# Global Variables
my $CWD = abs_path (dirname(__FILE__));
my $opt_max_datatype;
my $opt_update_files = "$CWD/update_files/";
my $opt_acc_output_file = "$CWD/acc_types.new";
my $opt_acc_input_file;
my $opt_classvar_output_file = "$CWD/classvar_mapping_config.new";
my $opt_classvar_input_file;
my $opt_overwrite;
my $opt_replace_origin;
my $opt_dryrun;
my $opt_help;
my $opt_debug;
my $opt_verbose;
my $opt_flood;

unless ( GetOptions("max_datatype=s"                    => \ $opt_max_datatype,
                    "update_files=s"                    => \ $opt_update_files,
                    "acc_input=s"                       => \ $opt_acc_input_file,
                    "acc_output=s"                      => \ $opt_acc_output_file,
                    "classvar_input=s"                  => \ $opt_classvar_input_file,
                    "classvar_output=s"                 => \ $opt_classvar_output_file,
                    "overwrite_existing_datatype"       => \ $opt_overwrite,
                    "replace_origin"                    => \ $opt_replace_origin,
                    "verbose"                           => \ $opt_verbose,
                    "debug"                             => \ $opt_debug,
                    "flood"                             => \ $opt_flood,
                    "help"                              => \ $opt_help )) {
    &print_usage;
}
if (defined $opt_help) {
	&print_usage;
}

&OpeniT::Defaults::log($opt_debug, $opt_verbose, "interactive");
my $CONFIG_DIR = &OpeniT::Config::CONFIG_DIR();
DEBUG("Root Dir: $CWD");

# Get acc_type input
if (! defined $opt_acc_input_file) {
    $opt_acc_input_file = $CONFIG_DIR . "/acc_types";
} 

my @data_types = &OpeniT::File::read_config_array_file($opt_acc_input_file) or 
    die "_main: Failed to read $opt_acc_input_file $!";
INFO ( "Detected " . $#data_types . " datatypes on installation" );

# Get update files
my %accupdates = _getAccUpdates($opt_update_files);

# Get the max datatype id update
if (!defined $opt_max_datatype ) {
    my $max_update;
    foreach (sort { $a <=> $b } keys(%accupdates) ) { $max_update = $_}
    $opt_max_datatype = $max_update;
}

# Add datatype if it does not exist.
my $start = ($#data_types + 1);
if ($start > $opt_max_datatype) {
    INFO( "Already reach max datatype id $opt_max_datatype. No new datatypes will be added.");
}
else {
    INFO( "Update will be from datatype $start to $opt_max_datatype ");
}

my $missing_updates = 0;
for my $ind ( $start .. $opt_max_datatype ) {

    if (!exists($accupdates{$ind})) {
        WARNING ( "Update file $ind.acc does not exists. " );
        $missing_updates += 1;
		next;
    }
    INFO( "Datatype $ind:'$accupdates{$ind}[0]' has been to acc_types" );
    push @data_types, $accupdates{$ind}; 
}

if ($missing_updates) {
	ERROR( "Detected $missing_updates missing update files. Exiting...." );
	exit -1;
}

# Actual writing of output
DEBUG ( "Target output file for acc_types $opt_acc_output_file " );
open(my $fh, '>', $opt_acc_output_file );
my $ind = 0;

print $fh "(\n";
foreach my $data_type (@data_types) {

    my $data_type_name  = $$data_type[0];
    my $measurements    = $$data_type[1];
    my $classifications = $$data_type[2];

    # Overwrite operations
    if ( defined $opt_overwrite && exists($accupdates{$ind}) )
    {
        $data_type_name  = $accupdates{$ind}[0];
        $measurements    = $accupdates{$ind}[1] ;
        $classifications = $accupdates{$ind}[2] ;
        INFO ( "Overwriting datatype $ind:$data_type_name" );
    }

    # Beautify output
    $data_type_name =~ tr/' '/'^'/;
    my $measure = _generateMeasurements($measurements);
    my $class   = _generateClassifications($classifications);

    # Write output
    ++$ind;
    print $fh "($data_type_name,\n";
    print $fh "$measure,\n";
    print $fh "$class)";  
    if ( $ind < scalar @data_types ) {
        print $fh ",\n\n";
    }
}
print $fh "\n\n)";
close $fh;
INFO( "Done writing new acc_type at '$opt_acc_output_file'");

if (defined $opt_replace_origin) {
    copy( $opt_acc_input_file ,"$opt_acc_input_file.old") or die "Failed to generate backup!$!";
    INFO( "Generated backup $opt_acc_input_file.old" );
    copy( $opt_acc_output_file ,"$opt_acc_input_file") or die "Failed replace origin file with output!$!";
    INFO( "Replaced origin file '$opt_acc_input_file' with '$opt_acc_output_file'" );
}

# Now were finish with acc_type lets proceed with classvar_mapping
INFO( "Running update on classvar_mapping");

if (! defined $opt_classvar_input_file) {
    $opt_classvar_input_file = $CONFIG_DIR . "/classvar_mapping_config";
} 

# Read classvar_mapping_config input 
my @classvars =  &OpeniT::File::read_config_array_file($opt_classvar_input_file) or 
    die "Failed to read $opt_classvar_input_file $!";

# Retrieve updates
my %classvar_updates = _getClassVarUpdates($opt_update_files);

if (defined $opt_flood) {
    print Dumper @classvars;
}

foreach my $key (keys %classvar_updates) {
    if (! _classvarExist ($key, \@classvars,) ) {
        INFO ("Adding new classvar '$key'");
        push @classvars, $classvar_updates{$key};
    } 
}




# Actual writing
open(my $fh, '>', $opt_classvar_output_file );
print $fh "(\n";
my $ind = 0;
foreach my $classvar (@classvars) { 
    my $classvar_name           = $$classvar[0];
    my $classvar_no_relation    = $$classvar[1];
    my $classvar_relation       = $$classvar[2];

    if (! &_accTypeExist($classvar_name, \@data_types) ) {
        &INFO("Classvar config '$classvar_name' does not exists on acc_types skipping...");
        next;
    }

    # Overwrite operations
    if (defined $opt_overwrite) {
        foreach (keys %classvar_updates) {

            if (lc $classvar_updates{$_}[0] eq lc $classvar_name ) {
                    &DEBUG("Classvar config $_ is already present in installation overwriting");
                    $classvar_name          = $classvar_updates{$_}[0];
                    $classvar_no_relation   = $classvar_updates{$_}[1] ;
                    $classvar_relation      = $classvar_updates{$_}[2] ;
                    last;
            }
        }
    }
    

    $classvar_name =~ tr/' '/'^'/;
    my $class_nr .= _generateClassVar($classvar_no_relation);
    my $class_r  .= _generateClassVar($classvar_relation);

    ++$ind;
    print $fh " ($classvar_name,\n";
    print $fh "  $class_nr,\n";
    print $fh "  $class_r)";  
    if ( $ind < scalar @classvars ) {
        print $fh ",\n\n";
    }

}

print $fh "\n\n)";
close $fh;

INFO( "Done with classvar_mapping_config.");

if (defined $opt_replace_origin) {
    copy( $opt_classvar_input_file ,"$opt_classvar_input_file.old") or die "Failed to generate backup!$!";
    INFO( "Generated backup $opt_classvar_input_file.old" );
    copy( $opt_classvar_output_file ,$opt_classvar_input_file) or die "Failed replace origin file with output!$!";
    INFO( "Replaced origin file '$opt_classvar_input_file' with '$opt_classvar_output_file'" );
}


sub _accTypeExist() {
    my ($name, $datatypes ) = @_;
    foreach (@$datatypes) {  
        my $data_type_name  = $$_[0];
        if ( lc $name eq lc $data_type_name) {
            return 1
        }
    }
    return 0;
}

sub _classvarExist() {
    my ($name, $classvars ) = @_;
    foreach my $classvar (@$classvars) {  
        my $classvar_name = $$classvar[0];
        if ( lc $name eq lc $classvar_name) {
            return 1
        }
    }
    return 0;
}

sub _generateClassVar() {
    my ($classvars) = @_;
    my $tmp = "(";
    my $ind = 0 ;

    for my $classvar (@$classvars)
    {
        if (ref($classvar) eq 'ARRAY' ) 
        {
            $tmp .= _generateClassVar($classvar);
        } else {
            $classvar =~ tr/' '/'^'/;
            $tmp .= $classvar;
        }

        $ind+=1;
        if ( $ind < scalar @$classvars ) {
            $tmp .= ", "
        }
        
    }
     $tmp .= ")";
    return $tmp;
}

sub _getClassVarUpdates() {
    
    my ($dir) = @_;
	chomp $dir;
	DEBUG( "Searching classvar files at $dir" );
    my @classvar_files = glob( $dir . '/*.classvar' );
    my %classvar_updates;
    for my $classvar_file (@classvar_files) {
        my @classvar_info = &OpeniT::File::read_config_array_file($classvar_file) or 
            die "Failed to read $classvar_file $!";
        my $classvar_name = $classvar_info[0];
        my $id = basename( $classvar_file ,  ".classvar");
        $classvar_updates{$classvar_name} = \@classvar_info;
        DEBUG( "Got update file $id:$classvar_name" );
    }
	
	if (defined $opt_flood ) {
		print Dumper %classvar_updates;
	}
	
    return %classvar_updates;
}

sub _getAccUpdates() {
    
    my ($dir) = @_;
	chomp $dir;
	DEBUG( "Searching acc files at $dir" );
    my @accfiles = glob( $dir . '/*.acc' );
    my %acc_updates;

    for my $acc (@accfiles) {
		DEBUG( "Reading update $acc" );
        my @accinfo = &OpeniT::File::read_config_array_file($acc) or 
            die "_getAccUpdates: Failed to read $acc! $!";
		my $id = basename( $acc ,  ".acc");
      
        if ( ! -e "$dir/$id.classvar") {
            WARNING(`Corresponding classvar update file for datatype $id not found!. 
            Please create update file $id.classvar`);
        }
        $acc_updates{$id} = \@accinfo;
    }
	
	if (defined $opt_flood ) {
		print Dumper %acc_updates;
	}
	
    return %acc_updates;
}

sub _generateClassifications() {
    my ($classifications) = @_;
    my $tmp = "(";
    my $ind = 0 ;

    for my $classification (@$classifications)
    {
        if (ref($classification) eq 'ARRAY' ) 
        {
            $tmp .= _generateClassifications($classification);
        } else {
            $classification =~ tr/' '/'^'/;
            $tmp .= $classification;
        }

        $ind+=1;
        if ( $ind < scalar @$classifications ) {
            $tmp .= ",\n  "
        }
        
    }
     $tmp .= ")";
    return $tmp;
}

sub _generateMeasurements() {
    my ($measurements) = @_;
    my $tmp = "(";
    my $ind = 0 ;

    for my $measurement (@$measurements)
    {
        
        if (ref($measurement) eq 'ARRAY' ) 
        {
            $tmp .= "\n  " . _generateMeasurements($measurement);
        } else {
                
            $measurement =~ tr/' '/'^'/;
            $tmp .= $measurement;
        }

        $ind+=1;
        if ( $ind < scalar @$measurements ) {
            $tmp  =  $tmp . ", ";
        } 
    }
    $tmp .= ")";
    return $tmp;
}

sub INFO() {
    my ($msg) = @_;
    print (" -- $msg \n");
}

sub print_usage {
my $CONFIG_DIR = &OpeniT::Config::CONFIG_DIR();
die "usage: " . &PROGNAME . " [options]

   A simple script for adding and updating `acc_types` and `classvar_mapping_config` 

  Options are:

    --update_files <dir>
          Directory where *.acc and *.classvar files is stored.
          Default: $CWD/update_files/
    --overwrite_existing_datatype
          Overwrite existing acc_types or classvar mapping configuration
    --replace_origin
          Replace '$CONFIG_DIR/acc_type' with output file. 
    --dryrun
          Do not replace '$CONFIG_DIR/acc_type' and $CONFIG_DIR/classvar_mapping_config file
          '$CWD/acc_type.new'
    --acc_input
          acc_type location. Default '$CONFIG_DIR/acc_type'
    --acc_output
          File path to save. Default '$CWD/acc_type.new'
    --classvar_input
          classvar_mapping_config location. 
          Default '$CONFIG_DIR/classvar_mapping_config'
    --classvar_output
          File path to save. Default '$CWD/classvar_mapping_config.new'
    --debug
          Write debug information.
    --verbose
          Be verbose.
    --flood
          Use for debugging.
    --help
          Displays this message.
";
}

### NOTE: DO NOT EXCEED 500 lines!!!!!